@@ -7,35 +7,89 @@ Wilson Júnior (wilsonpjunior@gmail.com).
 
 from django import forms
 from django.core.validators import EMPTY_VALUES
-from django.utils.encoding import smart_unicode
+try:
+    from django.utils.encoding import smart_text as smart_unicode
+except ImportError:
+    try:
+        from django.utils.encoding import smart_unicode
+    except ImportError:
+        from django.forms.util import smart_unicode
 from django.db.models.options import get_verbose_name
 from django.utils.text import capfirst
 
-from mongoengine import ReferenceField as MongoReferenceField
+from mongoengine import ReferenceField as MongoReferenceField, EmbeddedDocumentField as MongoEmbeddedDocumentField
 
-from .fields import MongoCharField, ReferenceField, DocumentMultipleChoiceField
+from .fields import MongoCharField, ReferenceField, DocumentMultipleChoiceField, ListField, MapField
 
 BLANK_CHOICE_DASH = [("", "---------")]
 
 class MongoFormFieldGenerator(object):
     """This class generates Django form-fields for mongoengine-fields."""
+    
+    # used for fields that fit in one of the generate functions
+    # but don't actually have the name.
+    generator_map = {
+        'sortedlistfield': 'generate_listfield',
+    }
+    
+    form_field_map = {
+        'stringfield': MongoCharField,
+        'stringfield_choices': forms.TypedChoiceField,
+        'stringfield_long': MongoCharField,
+        'emailfield': forms.EmailField,
+        'urlfield': forms.URLField,
+        'intfield': forms.IntegerField,
+        'intfield_choices': forms.TypedChoiceField,
+        'floatfield': forms.FloatField,
+        'decimalfield': forms.DecimalField,
+        'booleanfield': forms.BooleanField,
+        'booleanfield_choices': forms.TypedChoiceField,
+        'datetimefield': forms.DateTimeField,
+        'referencefield': ReferenceField,
+        'listfield': ListField,
+        'listfield_choices': forms.MultipleChoiceField,
+        'listfield_references': DocumentMultipleChoiceField,
+        'mapfield': MapField,
+        'filefield': forms.FileField,
+        'imagefield': forms.ImageField,
+    }
+    
+    # uses the same keys as form_field_map
+    widget_override_map = {
+        'stringfield_long': forms.Textarea,
+    }
+    
+    def __init__(self, field_overrides={}, widget_overrides={}):
+        self.form_field_map.update(field_overrides)
+        self.widget_override_map.update(widget_overrides)
 
     def generate(self, field, **kwargs):
         """Tries to lookup a matching formfield generator (lowercase
         field-classname) and raises a NotImplementedError of no generator
         can be found.
         """
+        # do not handle embedded documents here. They're more or less special
+        # and require some form of inline formset or something more complex
+        # to handle then a simple field
+        if isinstance(field, MongoEmbeddedDocumentField):
+            return
+        
         field_name = field.__class__.__name__.lower()
-        if hasattr(self, 'generate_%s' % field_name):
+        try:
             return getattr(self, 'generate_%s' % field_name)(field, **kwargs)
+        except AttributeError:
+            pass
 
         for cls in field.__class__.__bases__:
             cls_name = cls.__name__.lower()
-            if hasattr(self, 'generate_%s' % cls_name):
+            try:
                 return getattr(self, 'generate_%s' % cls_name)(field, **kwargs)
-
-        raise NotImplementedError('%s is not supported by MongoForm' % \
-                                          field.__class__.__name__)
+            except AttributeError:
+                if cls_name in self.form_field_map:
+                    return getattr(self, self.generator_map.get(cls_name))(field, **kwargs)
+                
+                raise NotImplementedError('%s is not supported by MongoForm' % \
+                                field.__class__.__name__)
 
     def get_field_choices(self, field, include_blank=True,
                           blank_choice=BLANK_CHOICE_DASH):
@@ -60,40 +114,58 @@ class MongoFormFieldGenerator(object):
     def get_field_label(self, field):
         if field.verbose_name:
             return field.verbose_name
-        return capfirst(get_verbose_name(field.name))
+        if field.name is not None:
+            return capfirst(get_verbose_name(field.name))
+        return ''
 
     def get_field_help_text(self, field):
         if field.help_text:
             return field.help_text.capitalize()
+        
+    def _check_widget(self, map_key):
+        if map_key in self.widget_override_map:
+            return {'widget': self.widget_override_map.get(map_key)}
+        else:
+            return {}
 
     def generate_stringfield(self, field, **kwargs):
-        form_class = MongoCharField
-
-        defaults = {'label': self.get_field_label(field),
-                    'initial': field.default,
-                    'required': field.required,
-                    'help_text': self.get_field_help_text(field)}
-
-        if field.max_length and not field.choices:
-            defaults['max_length'] = field.max_length
-
-        if field.max_length is None and not field.choices:
-            defaults['widget'] = forms.Textarea
-
-        if field.regex:
-            defaults['regex'] = field.regex
-        elif field.choices:
-            form_class = forms.TypedChoiceField
-            defaults['choices'] = self.get_field_choices(field)
-            defaults['coerce'] = self.string_field
-
-            if not field.required:
-                defaults['empty_value'] = None
-
+        if field.choices:
+            map_key = 'stringfield_choices'
+            defaults = {
+                'label': self.get_field_label(field),
+                'initial': field.default,
+                'required': field.required,
+                'help_text': self.get_field_help_text(field),
+                'choices': self.get_field_choices(field),
+                'coerce': self.string_field,
+            }
+        elif field.max_length is None:
+            map_key = 'stringfield_long'
+            defaults = {
+                'label': self.get_field_label(field),
+                'initial': field.default,
+                'required': field.required,
+                'help_text': self.get_field_help_text(field)
+            }
+        else:
+            map_key = 'stringfield'
+            defaults = {
+                'label': self.get_field_label(field),
+                'initial': field.default,
+                'required': field.required,
+                'help_text': self.get_field_help_text(field),
+                'max_length': field.max_length,
+            }
+            if field.regex:
+                defaults['regex'] = field.regex
+            
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
         return form_class(**defaults)
 
     def generate_emailfield(self, field, **kwargs):
+        map_key = 'emailfield'
         defaults = {
             'required': field.required,
             'min_length': field.min_length,
@@ -102,11 +174,13 @@ class MongoFormFieldGenerator(object):
             'label': self.get_field_label(field),
             'help_text': self.get_field_help_text(field)
         }
-
+        defaults.update(self._check_widget(map_key))
+        form_class = self.form_field_map.get(map_key)
         defaults.update(kwargs)
-        return forms.EmailField(**defaults)
+        return form_class(**defaults)
 
     def generate_urlfield(self, field, **kwargs):
+        map_key = 'urlfield'
         defaults = {
             'required': field.required,
             'min_length': field.min_length,
@@ -115,12 +189,14 @@ class MongoFormFieldGenerator(object):
             'label': self.get_field_label(field),
             'help_text':  self.get_field_help_text(field)
         }
-
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
-        return forms.URLField(**defaults)
+        return form_class(**defaults)
 
     def generate_intfield(self, field, **kwargs):
         if field.choices:
+            map_key = 'intfield_choices'
             defaults = {
                 'coerce': self.integer_field,
                 'empty_value': None,
@@ -130,10 +206,8 @@ class MongoFormFieldGenerator(object):
                 'choices': self.get_field_choices(field),
                 'help_text': self.get_field_help_text(field)
             }
-
-            defaults.update(kwargs)
-            return forms.TypedChoiceField(**defaults)
         else:
+            map_key = 'intfield'
             defaults = {
                 'required': field.required,
                 'min_value': field.min_value,
@@ -142,38 +216,44 @@ class MongoFormFieldGenerator(object):
                 'label': self.get_field_label(field),
                 'help_text': self.get_field_help_text(field)
             }
-
-            defaults.update(kwargs)
-            return forms.IntegerField(**defaults)
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
+        defaults.update(kwargs)
+        return form_class(**defaults)
 
     def generate_floatfield(self, field, **kwargs):
-
-        form_class = forms.FloatField
-
-        defaults = {'label': self.get_field_label(field),
-                    'initial': field.default,
-                    'required': field.required,
-                    'min_value': field.min_value,
-                    'max_value': field.max_value,
-                    'help_text': self.get_field_help_text(field)}
-
+        map_key = 'floatfield'
+        defaults = {
+            'label': self.get_field_label(field),
+            'initial': field.default,
+            'required': field.required,
+            'min_value': field.min_value,
+            'max_value': field.max_value,
+            'help_text': self.get_field_help_text(field)
+        }
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
         return form_class(**defaults)
 
     def generate_decimalfield(self, field, **kwargs):
-        form_class = forms.DecimalField
-        defaults = {'label': self.get_field_label(field),
-                    'initial': field.default,
-                    'required': field.required,
-                    'min_value': field.min_value,
-                    'max_value': field.max_value,
-                    'help_text': self.get_field_help_text(field)}
-
+        map_key = 'decimalfield'
+        defaults = {
+            'label': self.get_field_label(field),
+            'initial': field.default,
+            'required': field.required,
+            'min_value': field.min_value,
+            'max_value': field.max_value,
+            'help_text': self.get_field_help_text(field)
+        }
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
         return form_class(**defaults)
 
     def generate_booleanfield(self, field, **kwargs):
         if field.choices:
+            map_key = 'booleanfield_choices'
             defaults = {
                 'coerce': self.boolean_field,
                 'empty_value': None,
@@ -183,42 +263,51 @@ class MongoFormFieldGenerator(object):
                 'choices': self.get_field_choices(field),
                 'help_text': self.get_field_help_text(field)
             }
-
-            defaults.update(kwargs)
-            return forms.TypedChoiceField(**defaults)
         else:
+            map_key = 'booleanfield'
             defaults = {
                 'required': field.required,
                 'initial': field.default,
                 'label': self.get_field_label(field),
                 'help_text': self.get_field_help_text(field)
-                }
-
-            defaults.update(kwargs)
-            return forms.BooleanField(**defaults)
+            }
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
+        defaults.update(kwargs)
+        return form_class(**defaults)
 
     def generate_datetimefield(self, field, **kwargs):
+        map_key = 'datetimefield'
         defaults = {
             'required': field.required,
             'initial': field.default,
             'label': self.get_field_label(field),
         }
-
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
-        return forms.DateTimeField(**defaults)
+        return form_class(**defaults)
 
     def generate_referencefield(self, field, **kwargs):
+        map_key = 'referencefield'
         defaults = {
             'label': self.get_field_label(field),
             'help_text': self.get_field_help_text(field),
-            'required': field.required
+            'required': field.required,
+            'queryset': field.document_type.objects,
         }
-
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
-        return ReferenceField(field.document_type.objects, **defaults)
+        return form_class(**defaults)
 
     def generate_listfield(self, field, **kwargs):
+        # we can't really handle embedded documents here. So we just ignore them
+        if isinstance(field.field, MongoEmbeddedDocumentField):
+            return
+        
         if field.field.choices:
+            map_key = 'listfield_choices'
             defaults = {
                 'choices': field.field.choices,
                 'required': field.required,
@@ -226,39 +315,67 @@ class MongoFormFieldGenerator(object):
                 'help_text': self.get_field_help_text(field),
                 'widget': forms.CheckboxSelectMultiple
             }
-
-            defaults.update(kwargs)
-            return forms.MultipleChoiceField(**defaults)
         elif isinstance(field.field, MongoReferenceField):
+            map_key = 'listfield_references'
             defaults = {
                 'label': self.get_field_label(field),
                 'help_text': self.get_field_help_text(field),
-                'required': field.required
+                'required': field.required,
+                'queryset': field.document_type.objects,
             }
-
-            defaults.update(kwargs)
-            f = DocumentMultipleChoiceField(field.field.document_type.objects, **defaults)
-            return f
+        else:
+            map_key = 'listfield'
+            form_field = self.generate(field.field)
+            defaults = {
+                'label': self.get_field_label(field),
+                'help_text': self.get_field_help_text(field),
+                'required': field.required,
+                'field_type': form_field.__class__,
+            }
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
+        defaults.update(kwargs)
+        return form_class(**defaults)
+        
+    def generate_mapfield(self, field, **kwargs):
+        map_key = 'mapfield'
+        form_field = self.generate(field.field)
+        defaults = {
+            'label': self.get_field_label(field),
+            'help_text': self.get_field_help_text(field),
+            'required': field.required,
+            'field_type': form_field.__class__,
+        }
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
+        defaults.update(kwargs)
+        return form_class(**defaults)
 
     def generate_filefield(self, field, **kwargs):
+        map_key = 'filefield'
         defaults = {
             'required':field.required,
             'label':self.get_field_label(field),
             'initial': field.default,
             'help_text': self.get_field_help_text(field)
         }
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
-        return forms.FileField(**defaults)
+        return form_class(**defaults)
 
     def generate_imagefield(self, field, **kwargs):
+        map_key = 'imagefield'
         defaults = {
             'required':field.required,
             'label':self.get_field_label(field),
             'initial': field.default,
             'help_text': self.get_field_help_text(field)
         }
+        form_class = self.form_field_map.get(map_key)
+        defaults.update(self._check_widget(map_key))
         defaults.update(kwargs)
-        return forms.ImageField(**defaults)
+        return form_class(**defaults)
 
 
 class MongoDefaultFormFieldGenerator(MongoFormFieldGenerator):
